﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------

-- local cache 

local UnitGUID             = UnitGUID;
local UnitExists           = UnitExists;
local UnitIsPlayer         = UnitIsPlayer;
local UnitName             = UnitName;
local UnitIsUnit           = UnitIsUnit;
local UnitPlayerControlled = UnitPlayerControlled;
local GetTime              = GetTime;
local InCombatLockdown     = InCombatLockdown;

-------------------------------------------------------------------------------

nUI_ThreatSummary =
{
	self_agro   = false,	-- true if the player has agro
	pet_agro    = false,	-- true if the player pet has agro
	agro_name   = nil,		-- name of the mob's target
	max_name    = nil,		-- name of the player with the maximum threat
	next_name   = nil,		-- name of player with  next highest threat to agro
	max_guid    = nil,		-- GUID of the unit with the highest threat
	next_guid   = nil,		-- GUID of the unit with the second highest threat
	is_tanking  = false,	-- true if player is tanking
	my_threat   = 0,		-- the player's threat
	pet_threat  = 0,		-- the player's pet's threat
	max_threat  = 0,		-- the highest threat level
	next_threat = 0,		-- next highest threat to the agro threat
	agro_threat = 0,		-- the threat of the current agro target
	agro_gain   = 0,		-- the amount of agro needed to pull agro
	tank_regain = 0,		-- the max agro to recover a threat wipe
};

-- Ultimately, I would like that I wasn't doing any threat calculation at all
-- as the point is to integrate with other mods that already do this well.
-- Life is never that easy. KLH Threat Meter, once you wrap your head around
-- its abstraction, is easy enough to hit up for threat data. Sadly, Omen
-- leaves pretty much nothing of value as publicly accessible data, do I'm 
-- forced to just make use of the Ace Threat 2.0 library when present, which
-- is okay though it kind of makes me more responsible for calculating threat
-- which I loath to do. Such is the way of things

-------------------------------------------------------------------------------

nUI.Class.Threat = nUI.Class:CreatePrototype();
local prototype  = nUI.Class.Threat.prototype;

-------------------------------------------------------------------------------

function prototype:init( parent, id )
	
	self.parent        = parent;
	self.id            = id;
	self.timer         = 0;
	self.my_guid       = UnitGUID( "player" );
	self.pet_guid      = UnitExists( "pet" ) and UnitGUID( "pet" ) or nil;
	self.Callback      = {};
	self.ThreatTable   = {};
	self.ThreatSummary = {};
	
	if not self.event_frame then
		self.event_frame = CreateFrame( "Frame", "$parent_ThreatEvents", self.parent );
	end
	
	if nUI.AddOn.threat then
		self.active = true;
		self.Threat = LibStub( "Threat-2.0" );
	end
	
	if nUI.AddOn.ktm then
		self.active = true;
		self.KTM    = klhtm.threatlist.data;
	end
	
	if self.active then self:registerEvents(); end
	
end

-------------------------------------------------------------------------------

function prototype:registerEvents()
	
	self.event_frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.event_frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elpased ); end );
	
	self.event_frame:RegisterEvent( "PLAYER_TARGET_CHANGED" );
	self.event_frame:RegisterEvent( "PLAYER_REGEN_ENABLED" );
	self.event_frame:RegisterEvent( "PLAYER_REGEN_DISABLED" );
	self.event_frame:RegisterEvent( "PET_ATTACK_START" );
	self.event_frame:RegisterEvent( "PET_ATTACK_STOP" );
	self.event_frame:RegisterEvent( "UNIT_TARGET" );
	self.event_frame:RegisterEvent( "UNIT_PET" );

	if self.Threat then
		self.Threat.RegisterCallback( self, "ThreatUpdated" );
	end
end

-------------------------------------------------------------------------------

function prototype:unregisterEvents()
	
	self.event_frame:SetScript( "OnEvent", nil );
	self.event_frame:SetScript( "OnUpdate", nil );
	
	self.event_frame:UnregisterEvent( "PLAYER_TARGET_CHANGED" );
	self.event_frame:UnregisterEvent( "PLAYER_REGEN_ENABLED" );
	self.event_frame:UnregisterEvent( "PLAYER_REGEN_DISABLED" );
	self.event_frame:UnregisterEvent( "PET_ATTACK_START" );
	self.event_frame:UnregisterEvent( "PET_ATTACK_STOP" );
	self.event_frame:UnregisterEvent( "UNIT_TARGET" );
	self.event_frame:UnregisterEvent( "UNIT_PET" );
	
	if self.Threat then
		self.Threat.UnregisterCallback( self, "ThreatUpdated" );
	end
	
end

-------------------------------------------------------------------------------
-- this method is called from within Threat-2.0 when there's a change in any
-- unit's threat level

function prototype:ThreatUpdated( event, unit_guid, target_guid, threat )
	
	if target_guid == self.mob_guid then
		
		local data_set = self.ThreatTable[unit_guid] or {};
		
		if not data_set.threat then
			self.ThreatTable[unit_guid] = data_set;
		end
		
		data_set.threat = threat;
		data_set.when   = GetTime();
		
		self:updateThreatSummary();
		
	end
end

-------------------------------------------------------------------------------

function prototype:updateThreatSummary()

	local agro_exists = UnitExists( self.agro );
	local old_summary = self.ThreatSummary;
	local new_summary = 
	{
		self_agro   = agro_exists and UnitIsUnit( "player", self.agro ) or false,
		pet_agro    = agro_exists and UnitIsUnit( "pet", self.agro ) or false,
		agro_name   = agro_exists and UnitName( self.agro ) or nil,
		max_name    = nil,
		next_name   = nil,
		is_tanking  = self.Threat:GetModule( "Player" ).isTanking,
		my_threat   = 0,
		pet_threat  = 0,
		max_threat  = 0,
		next_threat = 0,
		agro_threat = 0,
		agro_gain   = 0,
		tank_regain = 0,
	};

	-- construct a new agro summary table from the existing threat table
	
	for i in pairs( self.ThreatTable ) do
		
		local data_set = self.ThreatTable[i];
		
		new_summary.agro_guid = self.agro_guid;

		if data_set.threat > new_summary.max_threat and i ~= self.pet_guid then
			new_summary.next_threat = new_summary.max_threat;
			new_summary.next_name   = new_summary.max_name;
			new_summary.next_guid   = new_summary.max_guid;
			new_summary.max_threat  = data_set.threat;
			new_summary.max_name    = self.Threat.GUIDNameLookup[i];
			new_summary.max_guid    = i;
		end
		
		if i == self.agro_guid then
			new_summary.agro_threat = data_set.threat;
		end
		
		if i == self.my_guid then
			new_summary.my_threat = data_set.threat;
		end
		
		if i == self.pet_guid then
			new_summary.pet_threat = data_set.threat;
		end
		
		if self.Threat:UnitInMeleeRange( self.mob ) then
			new_summary.agro_gain = new_summary.agro_threat * 1.1;
		else
			new_summary.agro_gain = new_summary.agro_threat * 1.3;
		end		
	end

	-- has anything changed in the table?
	
	local changed =	new_summary.self_agro   ~= old_summary.self_agro
				or	new_summary.pet_agro    ~= old_summary.pet_agro
				or	new_summary.pet_threat  ~= old_summary.pet_threat
				or	new_summary.my_threat   ~= old_summary.my_threat
				or	new_summary.agro_threat ~= old_summary.agro_threat
				or	new_summary.max_threat  ~= old_summary.max_threat
				or	new_summary.next_threat ~= old_summary.next_threat
				or	new_summary.agro_gain   ~= old_summary.agro_gain
				or	new_summary.tank_regain ~= old_summary.tank_regain
				or	new_summary.is_tanking  ~= old_summary.is_tanking
				or	new_summary.agro_name   ~= old_summary.agro_name
				or	new_summary.max_name    ~= old_summary.max_name
				or	new_summary.next_name   ~= old_summary.next_name;
				
	-- if something changed, update the local summary and fire callbacks
	
	if changed then
		
		self.ThreatSummary = new_summary;
		self:execCallbacks( "threat", new_summary );
		
	end
end

-------------------------------------------------------------------------------
-- the callback registry allows the threat system to call intereted clients
-- when there's something to report so that the only polling required occurs
-- here in this module rather than in multiple interested client modules.
--
-- The data set passed to the registered callback method is either nil if the
-- player is no longer in combat (there is no threat to report) or an instance
-- of nUI_ThreatSummary described above. The first argument passed is the 
-- source of the data... "threat" or "ktm"

function prototype:addCallback( owner, method )
	
	self.Callback[owner] = method;
	
end

function prototype:removeCallback( owner )
	
	self.Callback[owner] = nil;
	
end

function prototype:execCallbacks( type, data_set )
	
	for i in pairs( self.Callback ) do

		local callback = self.Callback[i];
		
		if callback then callback( type, data_set ); end
		
	end
end

-------------------------------------------------------------------------------

function prototype:onEvent()
	
	if event == "UNIT_PET" and UnitIsUnit( arg1, "player" ) then
		
		local pet_guid = nil;
		
		if UnitExists( "pet" ) then
			pet_guid = UnitGUID( "pet" );
		end
		
		if pet_guid ~= self.pet_guid then
			self.pet_guid = pet_guid;
			self:newTarget();
		end
		
	elseif event == "PLAYER_REGEN_ENABLED" or (not InCombatLockdown() and event == "PET_ATTACK_STOP") then
		
		self:execCallbacks( "threat", nil );
		self:execCallbacks( "ktm", nil );

		self.ThreatTable   = {};
		self.ThreatSummary = {};

	elseif event == "PLAYER_REGEN_DISABLED" or event == "PET_ATTACK_START" 
		or event == "UNIT_TARGET" or event == "PLAYER_TARGET_CHANGED" then
		
		self:newTarget();
		
	end	
end

-------------------------------------------------------------------------------

function prototype:onUpdate( elapsed )
	
	local tot_guid = UnitExists( "targettarget" ) and UnitGUID( "targettarget" ) or nil;
	local tott_guid = UnitExists( "targettargettarget" ) and UnitGUID( "targettargettarget" ) or nil;
	
	if tot_guid ~= self.tot_guid or tott_guid ~= self.tott_guid then
		
		self.tot_guid  = tot_guid;
		self.tott_guid = tott_guid;
		self:newTarget();
		
	end
end

-------------------------------------------------------------------------------

function prototype:newTarget()
	
	local have_agro = self:updatePrincipals();

	if self.have_agro and not have_agro then

		self.have_agro     = false;
		self.ThreatTable   = {};
		self.ThreatSummary = {};
		self:execCallbacks( "threat", nil );
		self:execCallbacks( "ktm", nil );
		
	elseif have_agro then
		
		self.have_agro = true;

		if self.new_mob then
			
			self.new_mob  = false;
			self.new_agro = false;
			
			self.ThreatTable = {};
			
			for guid, threat in self.Threat:IterateGroupThreatForTarget( self.mob_guid ) do
				if threat > 0 then
					self.ThreatTable[guid] =
					{
						threat = threat,
						when   = GetTime(),
					};
				end
			end
			
			self:updateThreatSummary();
			
		elseif self.new_agro then
			
			self.new_agro = false;
			
			self:updateThreatSummary();
			
		end			
	end
end

-------------------------------------------------------------------------------
-- in short, this stub just figures out who the bad guy is and who has his agro

function prototype:updatePrincipals()

	local have_agro   = true;	
	local checkTarget = function( unit_id )
		
		return UnitExists( unit_id ) and not UnitIsPlayer( unit_id ) and not UnitPlayerControlled( unit_id );
		
	end;
	
	if checkTarget( "target" ) then
		self.mob  = "target";
		self.agro = "targettarget";
	elseif checkTarget( "targettarget" ) then
		self.mob  = "targettarget";
		self.agro = "targettargettarget";
	else
		self.mob  = nil;
		self.agro = nil;
		have_agro = false;
	end
	
	if have_agro then
		
		local mob_guid  = UnitExists( self.mob ) and UnitGUID( self.mob ) or nil;
		local agro_guid = UnitExists( self.agro ) and UnitGUID( self.agro ) or nil;
		
		if mob_guid ~= self.mob_guid then
			self.new_mob  = true;
			self.mob_guid = mob_guid;
		end
		
		if agro_guid ~= self.agro_guid then
			
			self.ThreatSummary.agro_name = agro_guid and UnitName( self.agro ) or nil;
			self.ThreatSummary.self_agro = UnitIsUnit( "player", self.agro );
			self.ThreatSummary.pet_agro  = UnitIsUnit( "pet", self.agro );
			
			self.new_agro  = true;
			self.agro_guid = agro_guid;
		end
	end
	
	return have_agro;
	
end
